import json
import os
import unittest
from math import floor
from multiprocessing import Process
from pathlib import Path
from shutil import copyfile

import numpy as np
import requests
from urllib3 import Retry

from ..base.accessors import GenericImageDataAccessor, InMemoryDataAccessor
from ..base.models import SemanticSegmentationModel, InstanceSegmentationModel

from ..base.accessors import generate_file_accessor

class TestServerBaseClass(unittest.TestCase):
    """
    Base class for unittests of API functionality.  Implements both server and clients for testing.
    """

    app_name = 'model_server.base.api:app'

    def setUp(self) -> None:
        import uvicorn
        host = '127.0.0.1'
        port = 5001

        self.server_process = Process(
            target=uvicorn.run,
            args=(self.app_name, ),
            kwargs={'host': host, 'port': port, 'log_level': 'critical'},
            daemon=True
        )
        self.uri = f'http://{host}:{port}/'
        self.server_process.start()

    def _get_sesh(self):
        sesh = requests.Session()
        retries = Retry(
            total=5,
            backoff_factor=0.1,
        )
        sesh.mount('http://', requests.adapters.HTTPAdapter(max_retries=retries))
        return sesh

    def _get(self, endpoint):
        return self._get_sesh().get(self.uri + endpoint)

    def _put(self, endpoint, query=None, body=None):
        return self._get_sesh().put(
            self.uri + endpoint,
            params=query,
            data=json.dumps(body)
        )

    def tearDown(self) -> None:
        self.server_process.terminate()
        self.server_process.join()

    def copy_input_file_to_server(self):
        resp = self._get('paths')
        pa = resp.json()['inbound_images']
        copyfile(
            self.input_data['path'],
            Path(pa) / self.input_data['name']
        )
        return self.input_data['name']

    def get_accessor(self, accessor_id, filename=None):
        resp = self._put(f'/accessors/write_to_file/{accessor_id}', query={'filename': filename})
        where_out = self._get('paths').json()['outbound_images']
        fp_out = (Path(where_out) / resp.json())
        self.assertTrue(fp_out.exists())
        return generate_file_accessor(fp_out)


def setup_test_data():
    """
    Look for test data, create test output directory, parse and return meta information
    :return:
        meta (dict) of test data and paths
    """

    def _winpath(f):
        if not isinstance(f, str):
            return f
        p = f.split('/')
        if len(p) > 1:
            p[1] = p[1] + ':'
            return '\\'.join(p[1:])
        else:
            return f

    # places to look for test data
    data_paths = [
        os.environ.get('UNITTEST_DATA_ROOT'),
        _winpath(os.environ.get('UNITTEST_DATA_ROOT')),
        Path.home() / 'model_server' / 'testing',
        os.getcwd(),
    ]
    root = None

    # look for first instance of summary.json
    for dp in data_paths:
        if dp is None:
            continue
        sf = (Path(dp) / 'summary.json')
        if sf.exists():
            with open(sf, 'r') as fh:
                meta = json.load(fh)
                root = dp
                break

    if root is None:
        raise Exception('Could not find test data, try setting environmental variable UNITTEST_DATA_ROOT.')

    meta['root'] = Path(root)
    meta['output_path'] = meta['root'] / 'test_output'
    meta['output_path'].mkdir(parents=True, exist_ok=True)

    # resolve relative paths
    def _resolve_paths(d):
        keys = list(d.keys())
        for k in keys:
            if k == 'name':
                d['path'] = meta['root'] / d['name']
            elif isinstance(d[k], dict):
                _resolve_paths(d[k])
    _resolve_paths(meta)

    return meta

# object containing test data paths and metadata, for import into unittest modules
meta = setup_test_data()


class DummySemanticSegmentationModel(SemanticSegmentationModel):

    model_id = 'dummy_make_white_square'

    def load(self):
        return True

    def infer(self, img: GenericImageDataAccessor) -> (GenericImageDataAccessor, dict):
        super().infer(img)
        w = img.shape_dict['X']
        h = img.shape_dict['Y']
        result = np.zeros([h, w], dtype='uint8')
        result[floor(0.25 * h) : floor(0.75 * h), floor(0.25 * w) : floor(0.75 * w)] = 255
        return InMemoryDataAccessor(data=result), {'success': True}

    def label_pixel_class(
            self, img: GenericImageDataAccessor, **kwargs) -> GenericImageDataAccessor:
        mask, _ = self.infer(img)
        return mask


class DummyInstanceSegmentationModel(InstanceSegmentationModel):

    model_id = 'dummy_pass_input_mask'

    def load(self):
        return True

    def infer(
            self, img: GenericImageDataAccessor, mask: GenericImageDataAccessor
    ) -> (GenericImageDataAccessor, dict):
        return img.__class__(
            (mask.data / mask.data.max()).astype('uint16')
        )

    def label_instance_class(
            self, img: GenericImageDataAccessor, mask: GenericImageDataAccessor, **kwargs
    ) -> GenericImageDataAccessor:
        """
        Returns a trivial segmentation, i.e. the input mask with value 1
        """
        super(DummyInstanceSegmentationModel, self).label_instance_class(img, mask, **kwargs)
        return self.infer(img, mask)
